// Uebung5 Betriebssysteme 2
// Matthias Jauernig, 2006
// Server-Programm

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdbool.h>
#include <signal.h>
#include <sys/wait.h>
#include <poll.h>

#define SOCKNAME "sockunix"
#define MAXCLIENTS 10

#define SENDACK  "a"
#define SENDNACK "d"

bool ende = false;
struct pollfd fds[MAXCLIENTS+1];

struct clients {
	char fname[100];	// Dateiname des Clients
	bool hasFile;		// schon Dateinamen gesendet?
};

void sigintfunc(int signr){
	switch(signr){
	case SIGINT:
		printf("SIGINT empfangen...\n");
		ende = true;
		close(fds[0].fd);
		fds[0].fd = -1; // keine weiteren Verbindungen erlauben
		break;
	}
}

int main(void){
	int sockd, sockd2, l, sl, rc, i, scount=0, timeout=1000;
	struct sockaddr_un srv; // Bereich fuer Adressinformation
	struct pollfd *fdsp;
	struct clients client[MAXCLIENTS];
	nfds_t nfds;
	char eab[100]; // E/A-Bereich
	FILE *fp;
	
	// Adressinformationen bereitstellen
	srv.sun_family = AF_UNIX;
	strcpy(srv.sun_path, SOCKNAME);	// File fuer Datenaustausch eintragen

	// Socket einrichten
	if((sockd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0){
		perror("socket");
		return 1;
	}

	// als Vorsichtsmanahme Socket-File loeschen - darf noch nicht existieren
	unlink(srv.sun_path);
	sl = sizeof(srv);
	if(bind(sockd, (struct sockaddr*)&srv, sl) < 0){ // Socketfile mit bind() anlegen
		perror("bind");
		return 2;
	}

	// Warteschlange fuer Clients anlegen
	if(listen(sockd, MAXCLIENTS) < 0){
		perror("listen");
		return 3;
	}

	// Deskriptorfeld vorbereiten
	nfds = MAXCLIENTS+1; // MAXCLIENTS Feldelemente zu prfen
	fdsp = fds;
	fdsp->fd = sockd;
	fdsp->events = POLLIN;	// fr accept()
	
	// Deskriptorfeld fr Clients
	for(i=0; i<nfds; i++){
		fdsp++;
		fdsp->fd = -1;
		fdsp->events = POLLOUT; // um Client Anmeldebesttigung zu senden
	}
	
	sigset(SIGINT, sigintfunc);	// SIGINT verarbeiten
	printf("Warte auf Client-Anforderungen...\n");
	do{
		if((rc = poll(fds, nfds, timeout))<0 && !ende){
			perror("poll");
			return 4;
		}
		if(rc==0){
			//printf("poll() timeout abgelaufen...\n");
		}
		if(rc>0 || ende){ // rc = Anzahl Deskriptoren, fr die ein Ereignis anliegt
			if(fds[0].fd > 0 && (fds[0].revents&POLLIN) != 0){
				// Clientprozess aus Warteschlange holen
				printf("Neuer Client hat sich angemeldet...\n");
				if((sockd2 = accept(sockd, 0, 0))<0 && !ende){
					perror("accept");
					return 5;
				}
				// suchen des nchsten freien Platzes fr den Client
				for(i=1; i<=MAXCLIENTS && fds[i].fd>-1; i++);
				// Client eintragen oder ablehnen
				if(i>MAXCLIENTS){ // Client hat keinen Platz mehr
					printf("Kann Client nicht mehr aufnehmen...\n");
					sprintf(eab, "%s", SENDNACK);
					if((l=write(sockd2, eab, 2)) < 0){
						perror("write NACK");
						return 6;
					}
				}
				else{ // Client aufnehmen
					printf("Neuen Client an Position %d aufgenommen...\n", i);
					sprintf(eab, "%s", SENDACK);
					if((l=write(sockd2, eab, 2)) < 0){
						perror("write ACK");
						return 7;
					}
					// Schalte auf Empfang
					fds[i].fd = sockd2;
					fds[i].events = POLLIN;
					scount++;
					client[i-1].hasFile = false; // noch keinen Dateinamen bermittelt
				}
			}
			
			fdsp = fds;
			// jeden angemeldeten Client auf Empfang prfen
			for(i=1; i<=MAXCLIENTS && scount>0; i++){
				fdsp++;
				// Empfangsprfung
				if(fdsp->fd>0 && (fdsp->revents&POLLIN) != 0){
					// lesen und ACK zurckschicken
					if((l = read(fdsp->fd, eab, sizeof(eab))) < 0){
						perror("read");
						return 8;
					}else if(l>0){ // wenn Daten empfangen
						if(!client[i-1].hasFile){ // Dateiname noch nicht empfangen
							client[i-1].hasFile = true;
							strcpy(client[i-1].fname, eab);
							printf("Datei des neuen Clients: %s\n", client[i-1].fname);
						}
						else{ // normale Daten empfangen
							if((fp = fopen(client[i-1].fname, "a")) < 0){
								perror("fopen");
								return 10;
							}
							// Nachricht in Datei schreiben
							if(fwrite(&eab, strlen(eab)+1, 1, fp) < 0){
								perror("fwrite");
								return 11;
							}
							printf("Empfangene Daten in '%s' geschrieben: %s", client[i-1].fname, eab);
							fclose(fp);
							// Byteanzahl als Quittung zurcksenden
							fdsp->events = POLLOUT;
							sprintf(eab, "%d", l);
							if((l=send(fdsp->fd, eab, strlen(eab)+1, 0)) < 0){
								perror("send");
								return 12;
							}
						}
					}else if(l==0){ // wenn letzte Nachricht -> Ende
						scount--;
						close(fdsp->fd);
						fdsp->fd = -1;
						printf("Client mit Datei '%s' abgemeldet...\n", client[i-1].fname);
						break;
					}
				}
				
				// Sendeoperation prfen
				if(fdsp->fd>0 && (fdsp->revents&POLLOUT) != 0){
					fdsp->events = POLLIN; // bergang auf Empfang
				}
			}
		}
	}while(!ende || scount>0);

	// Socket entfernen
	if(unlink(srv.sun_path) < 0){
		perror("unlink");
		return 14;
	}

	// sockd schlieen
	close(sockd);

	return 0;
}
